from typing import Any, Optional
from urllib.parse import urljoin

import aiohttp

from huum.exceptions import SafetyException
from huum.schemas import HuumStatusResponse
from huum.settings import settings

API_BASE = "https://api.huum.eu/action/home/"


class Huum:
    """
    Usage:
        # Usage with env vars
        huum = Huum()

        # Setting auth variables explicitly
        huum = Huum(username="foo", password="bar)

        # If you don't have an existing aiohttp session
        # then run `open_session()` after initilizing
        huum.open_session()

        # Turn on the sauna
        huum.turn_on(80)
    """

    min_temp = 40
    max_temp = 110

    session: aiohttp.ClientSession

    def __init__(
        self,
        username: Optional[str] = settings.huum_username,
        password: Optional[str] = settings.huum_password,
        session: Optional[aiohttp.ClientSession] = None,
    ) -> None:
        if session:
            self.session = session

        if not username or not password:
            raise ValueError(
                "No username or password provided either by the environment nor explicitly"
            )
        self.auth = aiohttp.BasicAuth(username, password)

    async def handle_response(self, response: aiohttp.ClientResponse) -> Any:
        """
        Handle a response and raise errors if status code is not 200

        Args:
            response: Response from the Huum API

        Returns:
            A response object
        """
        if response.status != 200:
            response_text = await response.text()
            raise aiohttp.ClientError(
                f"Request failed with status code {response.status}. {response_text}"
            )
        return await response.json()

    async def _check_door(self) -> None:
        """
        Check if the door is closed, if not, raise an exception
        """
        status = await self.status()
        if not status.door_closed:
            raise SafetyException("Can not start sauna when door is open")

    async def status(self) -> HuumStatusResponse:
        """
        Get the status of the Sauna

        Returns:
            A `HuumStatusResponse` from the Huum API
        """
        url = urljoin(API_BASE, "status")

        response = await self.session.get(url, auth=self.auth)
        json_data = await self.handle_response(response)

        return HuumStatusResponse(**json_data)

    async def turn_on(
        self, temperature: int, safety_override: bool = False
    ) -> HuumStatusResponse:
        """
        Turns on the sauna at a given temperature

        Args:
            temperature: Target temperature to set the sauna to
            safety_override: If False, check if door is close before turning on the sauna

        Returns:
            A `HuumStatusResponse` from the Huum API
        """
        if temperature not in range(self.min_temp, self.max_temp):
            raise ValueError(
                f"Temperature '{temperature}' must be between {self.min_temp}-{self.max_temp}"
            )

        if not safety_override:
            await self._check_door()

        url = urljoin(API_BASE, "start")
        data = {"targetTemperature": temperature}

        response = await self.session.post(url, json=data, auth=self.auth)
        json_data = await self.handle_response(response)

        return HuumStatusResponse(**json_data)

    async def set_temperature(
        self, temperature: int, safety_override: bool = False
    ) -> HuumStatusResponse:
        """
        Alias for turn_on as Huum does not expose an explicit "set_temperature" endpoint

        Implementation choice: Yes, aliasing can be done by simply asigning
        set_temperature = turn_on, however this will not create documentation,
        makes the code harder to read and is generally seen as non-pythonic.

        Args:
            temperature: Target temperature to set the sauna to
            safety_override: If False, check if door is close before turning on the sauna

        Returns:
            A `HuumStatusResponse` from the Huum API
        """
        return await self.turn_on(temperature, safety_override)

    async def turn_off(self) -> HuumStatusResponse:
        """
        Turns off the sauna

        Returns:
            A `HuumStatusResponse` from the Huum API

        """
        url = urljoin(API_BASE, "stop")

        response = await self.session.post(url, auth=self.auth)
        json_data = await self.handle_response(response)

        return HuumStatusResponse(**json_data)

    async def open_session(self) -> None:
        self.session = aiohttp.ClientSession()

    async def close_session(self) -> None:
        await self.session.close()
